package bdd.search.spider;

import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.BufferedOutputStream;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URL;
import bdd.util.FIFOQueue;
import bdd.search.EnginePrefs;
import bdd.search.Monitor;

/** Written by Tim Macinta 1997                           <br>
 *  Distributed under the GNU Public License
 *       (a copy of which is enclosed with the source).   <br>
 *                                                        <br> 
 *  The Indexer is a thread which can index URLs that have been
 *  cached using the URLStatus class.  Use the queueURL() method
 *  to add cached URLs to the Indexer's list of URLs.  Once the
 *  start() method is called, the Indexer will start processing
 *  URLs in its queue. More URLs can also be added after calling
 *  start, in fact this may be the best way to use the Indexer.
 *  Calling the stopWhenDone() method will cause the Indexer
 *  thread to stop as soon as its queue empties.
 */

public class Indexer extends Thread {

  File working_dir;              // a temporary working directory
  FIFOQueue q = new FIFOQueue(); // queue for cached URLs
  Object q_mutex = new Object(); // synchronization mutex
  boolean running = false;       // set to false when the Thread should stop
  Crawler crawler;               // the crawler that retrieves URLs
  EnginePrefs prefs;             // preferences
  boolean exit_when_done = false;// exit when done indexing?
  long total_bytes = 0;          // total number of bytes indexed
  
  // names for temporary files
  
  static final String TMP_NAME = "tmp.db";
  static final String TMP_NAME_2 = "tmp2.db";

  /** "working_dir" should be a directory that only this
   *  Indexer and a given Cralwer will be
   *  accessing.  This means that if several Indexers are running
   *  simultaneously, they should all be given different "working_dir"
   *  directories.  Also, no other threads should write to this
   *  directory (except for the selected Crawler).
   */
  public Indexer(File working_dir, Crawler crawler, EnginePrefs prefs) {
    this.working_dir = working_dir;
    this.crawler = crawler;
    this.prefs = prefs;
    cleanUp();  // remove all temporary files
  }

  /** Use this method to add a cached url to the Indexer. */
  public void queueURL(URLStatus url) {
    if (url == null) return;
    synchronized (q_mutex) {
      q.addElement(url);
    }
  }

  /** Starts the Indexer. */
  public void start() {
    running = true;
    super.start();
  }

  /** This is where the actual indexing is done. */
  public void run() {
    
    URLStatus url;
    while (q.hasMoreElements() || running) {
      if (q.hasMoreElements()) {
	synchronized (q_mutex) {
	  url = (URLStatus) q.readNextElement();
	}
	try {
	  total_bytes += url.getContentLength();
	  Monitor m = prefs.getMonitor();
	  if (m != null) m.bytesIndexed(total_bytes);
	  File db = new File(working_dir, TMP_NAME);
	  DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(db)));
	  url.dumpToDatabase(out);
	  out.flush();
	  out.close();
	  addNewURLs(url.getLinkExtractor());
	  synchronized (q_mutex) {
	    q.nextElement();
	  }
	  mergeDatabases(db);
	} catch (IOException e) {
	  e.printStackTrace();
	}
      } else {

	// Nothing in queue so sleep for a few seconds

	try {
	  Thread.sleep(5000);
	} catch (InterruptedException e) {}
      }
    }
    try {
      replaceMainIndex();  // replace main index
    } catch (IOException e) {
      e.printStackTrace();
    }
    System.gc();
    cleanUp();  // remove all temporary files
    if (exit_when_done) {
      System.exit(0);
    }
  }

  /** Causes this Indexer to stop whenever it finishes indexing the URLs
   *  in its queue.         */
  public void stopWhenDone(boolean exit_when_done) {
    running = false;
    this.exit_when_done = exit_when_done;
  }

  /** Removes all the ".db" and ".tmp" files in the directory "working_dir". */
  void cleanUp() {
    String[] files = working_dir.list();
    for (int i = 0; i < files.length; i++) {
      if (files[i].endsWith(".db") || files[i].endsWith(".tmp")) {
	new File(working_dir, files[i]).delete();
      }
    }
  }

  /** Repeatedly attempts to merge "temporary" with other temporary
   *  databases which have been merged the same number of times.  In other
   *  words, this method will first try to merge "temporary" with any
   *  databases that haven't been merged yet.  If that is successful,
   *  this database will then be merged with any databases that have been
   *  merged once.  If that is successful, this database will then be
   *  merged with any databases that have been merged twice... and
   *  so on and so forth.
   *  <p>
   *  Databases are named based on the number of times they have been
   *  merged.  E.g., a file called "6.db" will have been merged six times while
   *  a file called "9.db" will have been merged nine times.  It is assumed
   *  that the "temporary" file has not been merged at all.
   */
  void mergeDatabases(File temporary) throws IOException {
    int i = 0;
    File f = new File(working_dir, i+".db");
    File tmp2 = new File(working_dir, TMP_NAME_2);
    while (f.exists()) {
      merge(temporary, f, tmp2);
      temporary.delete();
      f.delete();
      tmp2.renameTo(temporary);
      i++;
      f = new File(working_dir, i+".db");
    }
    temporary.renameTo(f);
  }

  /** Adds new URLs to the crawler's queue.  */
  void addNewURLs(LinkExtractor urls) {
    while (urls.hasMoreElements()) {
      crawler.addURL((URL) urls.nextElement());
    }
  }

  /** Completes the merging of all temporary databases and replaces the
   *  main database with the final product.
   */
  void replaceMainIndex() throws IOException {

    // merge all existing databases

    String[] files = working_dir.list();
    int targ;
    boolean good;
    int count = 0;
    File tmp1 = new File(working_dir, TMP_NAME);
    File tmp2 = new File(working_dir, TMP_NAME_2);
    for (int i = files.length - 1; i >= 0; i--) {
      if (files[i].endsWith(".db")) {
	targ = files[i].length() - 3;
	if (targ > 0) {
	  good = true;
	  for (int x = 0; x < targ; x++) {
	    if (!Character.isDigit(files[i].charAt(x))) {
	      good = false;
	      break;
	    }
	  }
	  if (good) {
	    count++;
	    File f = new File(working_dir, files[i]);
	    merge(tmp1, f, tmp2);
	    tmp1.delete();
	    tmp2.renameTo(tmp1);
	    f.delete();
	  }
	}
      }
    }

    // replace main database

    if (count < 1) return;
    File main_index = prefs.getMainIndex();
    main_index.delete();
    tmp1.renameTo(main_index);
  }

  /** Takes two search databases, "file1" and "file2", and merges their
   *  contents with the results being placed in "target".  "file2" must
   *  exist, but "file1" need not.  If "file1" does not exist then
   *  "file2" is copied to "target".
   */
  void merge(File file1, File file2, File target) throws IOException {

    // copy file2 if file1 doesn't exist

    if (!file1.exists()) {
      OutputStream out =new BufferedOutputStream(new FileOutputStream(target));
      pipe(new FileInputStream(file2), out);
      out.flush();
      out.close();
      return;
    }
    
    // file1 does exist

    DataInputStream in1 = new DataInputStream(new FileInputStream(file1));
    DataInputStream in2 = new DataInputStream(new FileInputStream(file2));
    DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(target)));
    
    // merge headers

    out.writeLong(in1.readLong() + in2.readLong() -8 -4);  // pointer to words
    int offset1 = in1.readInt();
    int offset2 = in2.readInt();
    out.writeInt(offset1 + offset2);     // number of URLs

    // dump URL descriptions

    for (int i = 0; i < offset1; i++) {
      out.writeBytes(in1.readLine()+'\n');
    }
    for (int i = 0; i < offset2; i++) {
      out.writeBytes(in2.readLine()+'\n');
    }

    // dump actual words

    int url_num;
    String word1 = in1.readLine();
    String word2 = in2.readLine();
    while (word1 != null || word2 != null) {
      if (word1 == null) {

	// words from "file1" are done

	while (word2 != null) {
	  out.writeBytes(word2+'\n');
	  while (true) {
	    url_num = in2.readInt();
	    if (url_num == 0) break;
	    out.writeInt(offset1 + url_num);
	    out.write(in2.read());
	  }
	  out.writeInt(0);
	  word2 = in2.readLine();
	}
      } else if (word2 == null) {

	// words from "file2" are done
	
	out.writeBytes(word1+'\n');
	pipe(in1, out);
	word1 = null;
      } else {

	// still merging both files

	int compare = word1.compareTo(word2);
	if (compare == 0) {
	  
	  // same word in both databases

	  int u1 = in1.readInt();
	  int u2 = in2.readInt();
	  out.writeBytes(word1+'\n');
	  while (u1 != 0 || u2 != 0) {
	    if (u1 == 0) {
	      
	      // word1 is done

	      while (u2 != 0) {
		out.writeInt(u2 + offset1);
		out.write(in2.read());
		u2 = in2.readInt();
	      }
	    } else if (u2 == 0) {
	      
	      // word2 is done

	      while (u1 != 0) {
		out.writeInt(u1);
		out.write(in1.read());
		u1 = in1.readInt();
	      }
	    } else {

	      // neither is done yet

	      int s1 = ((int) in1.read()) & 0xff;
	      int s2 = ((int) in2.read()) & 0xff;
	      while (true) {
		if (s1 > s2) {
		
		  // the URL in file1 has a higher score

		  out.writeInt(u1);
		  out.write(s1);
		  u1 = in1.readInt();
		  if (u1 == 0) {
		    out.writeInt(u2 + offset1);
		    out.write(s2);
		    u2 = in2.readInt();
		    break;
		  } else {
		    s1 = ((int) in1.read()) & 0xff;
		  }
		} else {
		
		  // the URL in file2 has a higher score

		  out.writeInt(u2 + offset1);
		  out.write(s2);
		  u2 = in2.readInt();
		  if (u2 == 0) {
		    out.writeInt(u1);
		    out.write(s1);
		    u1 = in1.readInt();
		    break;
		  } else {
		    s2 = ((int) in2.read()) & 0xff;
		  }
		}
	      }
	    }
	  }
	  out.writeInt(0);
	  word1 = in1.readLine();
	  word2 = in2.readLine();
	} else if (compare > 0) {

	  // dump word2

	  out.writeBytes(word2+'\n');
	  while (true) {
	    url_num = in2.readInt();
	    if (url_num == 0) break;
	    out.writeInt(offset1 + url_num);
	    out.write(in2.read());
	  }
	  out.writeInt(0);
	  word2 = in2.readLine();
	} else {   // compare < 0

	  // dump word1

	  out.writeBytes(word1+'\n');
	  while (true) {
	    url_num = in1.readInt();
	    out.writeInt(url_num);
	    if (url_num == 0) break;
	    out.write(in1.read());
	  }
	  word1 = in1.readLine();
	}
      }
    }

    // close all files

    out.flush();
    in1.close();
    in2.close();
    out.close();
  }

  /** Pipes "in" to "out" until "in" is exhausted then closes "in".   */
  void pipe(InputStream in, OutputStream out) throws IOException {
    byte[] b = new byte[512];
    int x = in.read(b, 0, b.length);
    while (x > 0) {
      out.write(b, 0, x);
      x = in.read(b, 0, b.length);
    }
    in.close();
  }

}
