package bdd.util;

/** Written by Tim Macinta 1997                           <br>
 *  Distributed under the GNU Public License
 *       (a copy of which is enclosed with the source).   <br>
 *                                                        <br> 
 *  This class provides a simple First In First Out Queue.
 *  <b>Please Note:</b> this implementation of a FIFOQueue is not
 *  thread safe so if you plan on accessing the same queue with
 *  multiple threads you wil need to handle synchronization yourself
 *  (e.g., by creating a synchronized subclass).
 */

public class FIFOQueue implements java.util.Enumeration {
  int grow_size;
  int insert_point = 0;  // Point in array where next Object can be stored
  int read_point = 0;    // 
  Object[] q;            // Where the queued objects are stored

  public FIFOQueue() {
    this(15, 10);
  }

  /** "initial_cap" specifies the initial capacity of the queue and
   *  "grow_size" specifies the amount that the capacity grows by when
   *  capacity is reached.  Both parameters must be positive.
   */
  public FIFOQueue(int initial_cap, int grow_size) {
    this.grow_size = grow_size;
    q = new Object[initial_cap];
  }

  /** Adds "object" to the queue.   "object" CAN be null. */
  public void addElement(Object object) {
    if (insert_point == q.length) {  // make sure insertion point is valid
      if (read_point == 0) {
	grow();  // queue is full so grow it
      } else {
	shift(); // shift elements over to make room for insertion
      }
    }
    q[insert_point] = object;  // insert object
    insert_point++;
  }

  /** Enlarges the queue capacity by "grow_size". */
  void grow() {
    Object[] q2 = new Object[q.length + grow_size];
    System.arraycopy(q, 0, q2, 0, q.length);
    q = q2;
  }

  /** Shifts all elements so that "read_point" is set to 0. */
  void shift() {
    System.arraycopy(q, read_point, q, 0, insert_point - read_point);
    insert_point -= read_point;
    read_point = 0;
  }

  /** Removes the next element from the queue and returns it.
   *  <p>
   *  Throws a java.util.NoSuchElementException if the queue is empty.
   */
  public Object nextElement() {
    if (read_point == insert_point) {
      throw new java.util.NoSuchElementException("Queue is empty");
    }
    Object obj = q[read_point];
    q[read_point] = null;   // clear reference to allow for garbage collection
    read_point++;
    return obj;
  }

  /** Returns the next element that on the queue, but leaves the element
   *  on the queue so that the next call to nextElement() or
   *  readNextElement() will return the same element.
   *  <p>
   *  Throws a java.util.NoSuchElementException if the queue is empty.
   */
  public Object readNextElement() {
    if (read_point == insert_point) {
      throw new java.util.NoSuchElementException("Queue is empty");
    }
    return q[read_point];
  }

  /** Returns true if this queue contains more elements and false otherwise. */
  public boolean hasMoreElements() {
    return insert_point != read_point;
  }

  /** Returns the number of elements in this queue. */
  public int countElements() {
    return insert_point - read_point;
  }

  /** Conserves memory by shrinking the capacity of the queue to 
   *  "new_capacity".  Of course, you can also use this method to manually
   *  grow the size of the queue, but this is unecessary.
   *  "new_capacity" should be greater than or equal to
   *  the number of elements in the queue - if it is less than the
   *  number of elements in the queue, the capacity will be set to
   *  the number of elements in the queue.
   */
  public void setCapacity(int new_capacity) {
    int current_size = insert_point - read_point;
    if (new_capacity < current_size) new_capacity = current_size;
    Object q2 = new Object[new_capacity];
    System.arraycopy(q, read_point, q2, 0, current_size);
  }

  /** Sets the capacity to the number of elements in the queue.
   *  Equivalent to setCapacity(countElements()) .
   */
  public void shrinkCapacityToSize() {
    setCapacity(countElements());
  }

  /** Returns the element at index number "index" (index number 0 is
   *  the first element that is read with the nextElement() method).   */
  public Object readElementAt(int index) {
    if (index > insert_point - read_point) {
      throw new IndexOutOfBoundsException(index+" > "+(insert_point-read_point));
    } else if (index < 0) {
      throw new IndexOutOfBoundsException(index+" < "+0);
    }
    return q[index + read_point];
  }

  /** Insert "element" in the queue at index number "index."     */
  public void insertElementAt(Object element, int index) {
    if (insert_point == q.length) {  // make sure insertion point is valid
      if (read_point == 0) {
	grow();  // queue is full so grow it
      } else {
	shift(); // shift elements over to make room for insertion
      }
    }

    // shift all objects from "index" on to the right one`

    index += read_point;
    System.arraycopy(q, index, q, index+1, insert_point - index);
    q[index] = element;  // insert object
    insert_point++;
  }

}
